/**************************************************************
 * gsvoices.h  -  list of gs voice names  -  all 654 of SC88  *
 **************************************************************/
#ifndef GSVOICES_H
#define GSVOICES_H
char *gsvoice[128] = {
"Piano 1     ", "Piano 2     ", "Piano 3     ", "Honky-tonk  ",
"E.Piano 1   ", "E.Piano 2   ", "Harpsichord ", "Clav.       ",
"Celesta     ", "Glockenspl  ", "Music Box   ", "Vibraphone  ",
"Marimba     ", "Xylophone   ", "Tubularbell ", "Santur      ",
"Organ 1     ", "Organ 2     ", "Organ 3     ", "Church Org1 ",
"Reed Organ  ", "Accordion F ", "Harmonica   ", "Bandoneon   ",
"Nylon Gt.   ", "Steel Gt.   ", "Jazz Gt.    ", "Clean Gt.   ",
"Muted Gt.   ", "OverdriveGt ", "Dist.Gt.    ", "Gt.Harmonix ",
"Acoustic Bs ", "Fingered Bs ", "Picked Bass ", "Fretless Bs ",
"Slap Bass 1 ", "Slap Bass 2 ", "Syn.Bass 1  ", "Syn.Bass 2  ",
"Violin      ", "Viola       ", "Cello       ", "Contrabass  ",
"Tremolo Str ", "Pizzicato   ", "Harp        ", "Timpani     ",
"Strings     ", "SlowStrings ", "SynStrings1 ", "SynStrings2 ",
"Choir Aahs  ", "Voice Oohs  ", "SynVox      ", "Orchest.Hit ",
"Trumpet     ", "Trombone    ", "Tuba        ", "MuteTrumpet ",
"French Horn ", "Brass 1     ", "Syn.Brass 1 ", "Syn.Brass 2 ",
"Soprano Sax ", "Alto Sax    ", "Tenor Sax   ", "BaritoneSax ",
"Oboe        ", "EnglishHorn ", "Bassoon     ", "Clarinet    ",
"Piccolo     ", "Flute       ", "Recorder    ", "Pan Flute   ",
"Bottle Blow ", "Shakuhachi  ", "Whistle     ", "Ocarina     ",
"Square Wave ", "Saw Wave    ", "SynCalliope ", "ChifferLead ",
"Charang     ", "Solo Vox    ", "5th Saw     ", "Bass & Lead ",
"Fantasia    ", "Warm Pad    ", "Polysynth   ", "Space Voice ",
"Bowed Glass ", "Metal Pad   ", "Halo Pad    ", "Sweep Pad   ",
"Ice Rain    ", "Soundtrack  ", "Crystal     ", "Atmosphere  ",
"Brightness  ", "Goblin      ", "Echo Drops  ", "Star Theme  ",
"Sitar       ", "Banjo       ", "Shamisen    ", "Koto        ",
"Kalimba     ", "Bagpipe     ", "Fiddle      ", "Shanai      ",
"Tinkle Bell ", "Agogo       ", "Steel Drums ", "Woodblock   ",
"Taiko       ", "Melo. Tom 1 ", "Synth Drum  ", "Reverse Cym ",
"Gt.FretNoiz ", "BreathNoise ", "Seashore    ", "Bird        ",
"Telephone 1 ", "Helicopter  ", "Applause    ", "Gun Shot    " };
#endif
